import time
from typing import Any, Callable, Tuple


def reRunTimes(times: int = 1) -> Callable:
    """一个装饰器, 用来计算函数运行时长, 第一个装饰器参数是运行次数

    :param times: 运行次数, 默认为1
    :return: 装饰器函数
    """

    def ruturnFun(fun: Callable) -> Callable:
        """一个装饰器, 用来计算函数运行时长, 第一个装饰器参数是运行次数

        :param fun: 原函数
        :return: 装饰器函数
        """

        def runFun(*args, **kwargs) -> Tuple[Any, int]:
            """一个装饰器, 用来计算函数运行时长, 第一个装饰器参数是运行次数

            :return: 返回一个元组:(原函数调用结果, ns为单位的运行时长)
            """
            _time_start = time.perf_counter_ns()
            for _ in range(times):
                fun_ret = fun(*args, **kwargs)
            time_used = time.perf_counter_ns() - _time_start
            return fun_ret, time_used

        return runFun

    return ruturnFun


def funName(fun: Callable) -> Callable:
    """一个装饰器, 函数形参增加__fun_name__用于获得函数名
    ---------------------------------------------

    使用方法:

    原函数
        def a(c, b: int = 1):

    ->

    改成
        def a(c, b: int = 1, __fun_name__):


    :param fun: 原函数
    :return: 装饰器函数
    """
    name = fun.__name__

    def ret_fun(*args, **kwargs) -> Any:
        """装饰函数

        :return: 函数运算结果
        """
        return fun(*args, __fun_name__=name, **kwargs)

    return ret_fun



from typing import Union


def flatten(sequence: Union[list, tuple]) -> list:
    """将多维数据结构展平为一纬数据结构

    :param sequence: 多维数据结构
    :return: 一纬数据结构
    """
    ret = []
    for _ in sequence:
        if isinstance(_, list):
            ret.extend(flatten(_))
        else:
            ret.append(_)
    return ret


def flatten_layer(sequence: Union[list, tuple]) -> list:
    """将多维数据结构展平一层

    :param sequence: 多维数据结构
    :return: 展平了一层的一纬数据结构
    """
    ret = []
    for _ in sequence:
        if isinstance(_, list):
            ret.extend(_)
        else:
            ret.append(_)
    return ret


def flatten_no_recursion(sequence: Union[list, tuple]) -> list:
    """将多维数据结构展平为一纬数据结构(无递归)效率低很多

    :param sequence: 多维数据结构
    :return: 一纬数据结构
    """
    while True:
        if all(not isinstance(_, list) for _ in sequence):  # 全不是列表
            return sequence
        sequence = flatten_layer(sequence)



"""默认导入hpycore的所有属性和函数，易于使用"""
from .hpycore import *
from .hpysettings import SettingsFileObject

name = "hpyculator"
__all__ = ["hpy" + body for body in "core, decorator, settings, func".split(",")]



from __future__ import annotations

import os
from abc import ABC, abstractmethod
from typing import Any


class SettingsFileObject(ABC):
    """设置文件的基类，要新支持一种格式就继承这个抽象类"""

    def __init__(
        self,
        settings_dir_path: str,
        settings_file_name: str = "settings",
        settings_file_format: str = "",
    ):
        """读取一个文件, 初始化文件对象

        :param settings_dir_path: 设置文件所在的目录
        :type settings_dir_path: str
        :param settings_file_name: 设置文件的名称, 默认为"settings"
        :type settings_file_name: str
        :param settings_file_format: 设置文件的后缀, 默认为""
        :type settings_file_format: str
        """
        self._setting_dir_path = settings_dir_path
        # 检查存放设置文件的文件夹是否存在
        if not os.path.exists(self._setting_dir_path):
            os.makedirs(self._setting_dir_path)

        # 初始化设置文件位置
        self._settings_file_path = str(
            os.path.join(
                settings_dir_path, f"{settings_file_name}.{settings_file_format}"
            )
        )

        # 初始化文件
        self._settings_file_stream = open(
            self._settings_file_path, mode="a+", encoding="utf-8"
        )
        self._settings_file_stream.close()

    @abstractmethod
    def add(self, key: str, value: Any) -> SettingsFileObject:
        """添加一项配置

        :param key: 键
        :type key: str
        :param value: 值
        :type value: Any
        :return: 链式调用, 返回本身
        :rtype: SettingsFileObject
        """

    @abstractmethod
    def read(self, key: str) -> Any:
        """读取一项配置

        :param key: 键
        :type key: str
        :return: 值
        :rtype: Any
        :raises KeyError: 没有这个键
        """
        if not self.exists(key):
            raise KeyError

    @abstractmethod
    def readAll(self):
        """读取全部的

        :return: 键值对
        :rtype: dict
        """

    @abstractmethod
    def delete(self, key: str):
        """删除一项配置

        :param key: 键
        :type key: str
        :return: 链式调用, 返回本身
        :rtype: SettingsFileObject
        :raises KeyError: 没有这个键
        """
        if not self.exists(key):
            raise KeyError

    @abstractmethod
    def modify(self, key: str, value: Any):
        """修改一项配置

        :param key: 键
        :type key: str
        :param value: 值
        :type value: Any
        :return: 链式调用, 返回本身
        :rtype: SettingsFileObject
        :raises keyError: 没有这个键
        """
        if not self.exists(key):
            raise KeyError

    @abstractmethod
    def exists(self, key: str) -> bool:
        """检查一个键是否存在

        :param key: 键
        :type key: str
        :return: 存在为True, 不存在为False
        :rtype: bool
        """

    @property
    def setting_file_path(self) -> str:
        """设置文件的路径

        :return: 设置文件的路径
        :rtype: str
        """
        return self._settings_file_path



"""管理设置文件"""
# xml X
# ini X

import os
from typing import Union

from .json_file_object import JsonSettingsFileObject
from .toml_file_object import TomlSettingsFileObject
from .yaml_file_object import YamlSettingsFileObject

dict_settings_file_object = {
    "toml": TomlSettingsFileObject,
    "json": JsonSettingsFileObject,
    "yaml": YamlSettingsFileObject,
}

SettingsFileObject = Union[
    JsonSettingsFileObject, YamlSettingsFileObject, TomlSettingsFileObject
]


def load(
    settings_dir_path: str = str(os.path.join(os.getcwd(), "Setting")),  # 默认设置目录
    settings_file_name: str = "settings",
    settings_file_format: str = "toml",
) -> SettingsFileObject:
    """加载一个设置文件对象
    工厂函数

    :param settings_dir_path: 设置文件所在目录, 默认是 str(os.path.join(os.getcwd(), "Setting"))
    :param settings_file_name: 设置文件的文件名, 默认是"settings"
    :param settings_file_format: 设置文件的类型(后缀), 默认是"toml"
    :return: 设置文件实例
    """
    return dict_settings_file_object[settings_file_format](
        settings_dir_path, settings_file_name, settings_file_format
    )
    # settings_dir_path = self._settings_dir_path  # 易于语法高亮
    # return JsonSettingsFileObject(settings_dir_path,
    #                               settings_file_name,
    #                               settings_file_format)



from queue import Queue
from typing import IO, Any

io_instance = None  # 我想用类属性的，但是时间给我干到原来的三倍，傻眼了
# 不加type hint的原因https://github.com/python/cpython/issues/79120
# https://stackoverflow.com/questions/60466092/why-does-order-of-declaration-matter-for-annotated-global-variables-in-modules

STRING = 1 << 0
NUM = 1 << 1
FLOAT = 1 << 2
LIST = 1 << 3

ON = 1 << 1
OFF = 1 << 0  # 因为没读到就用这个作为缺省，所以每个参数的(1<<0)就是缺省

# 'output_mode'
RETURN_ONCE = 1 << 0
RETURN_ITERABLE = 1 << 1
# RETURN_ITERABLE_OUTPUT_IN_ONE_LINE = 1 << 2
NO_RETURN = 1 << 3
NO_RETURN_SINGLE_FUNCTION = 1 << 4

message_queue: Queue = Queue()  # 输出消息 保证里面取出来的数据类型一定是str
# OutputReachedLimit
# CalculationProgramIsRunning
# CalculationProgramIsFinished（下一条消息是所用时间，类型int，单位ns）
# TypeConversionError
# CalculationError


def write(anything: Any, end: str = "\n") -> None:
    """
    用于向指定的文件流写入，每次写入之后立即刷新缓存区（立即写入硬盘）

    :param anything: 要写入的东西
    :param end: 每次写入在末尾追加的东西，默认为"\n"
    """
    io_instance.write(str(anything) + end)  # type: ignore
    io_instance.flush()  # type: ignore


def write_without_flush(anything: Any, end: str = "\n") -> None:
    """用于向指定的文件流写入, 每次写入之后不刷新缓存区, 需要手动刷新(使用flush函数)

    :param anything: 要写入的东西
    :param end: 每次写入在末尾追加的东西，默认为"\n"
    """
    io_instance.write(str(anything) + end)  # type: ignore


def flush() -> None:
    """用于刷新缓存区(将缓存区中的数据写入硬盘)"""
    io_instance.flush()  # type: ignore


def output(anything: Any) -> None:
    """输出到框体，会自动添加换行符

    :param anything: 要输出到框体的数据
    """
    if isinstance(anything, str):
        message_queue.put(("OUTPUT", anything))
    else:
        message_queue.put(("OUTPUT", str(anything)))


# 以下是用来传递数据的


def setIoInstance(instance: IO) -> None:
    """设置类属性: io实例

    :param instance: io实例
    """
    global io_instance
    io_instance = instance


def getIoInstance() -> IO:
    """返回io实例

    :return: io实例
    """
    return io_instance  # type: ignore  # 难以详细的标出IO类型


def addOne(num: int) -> int:
    """用于测试的函数，会输出输入数字+1的结果

    :param num: 一个数字
    :return: 输入+1
    """
    return num + 1



from random import randint

from . import hpydecorator

# 测试数据
num = randint(0, 99999)
test_data = str(num) + "test_data"


def test_hpydecorator():
    test_times: int = 0  # 初始化一个变量，用于检测结果

    @hpydecorator.reRunTimes(5)
    def _fun(text: str):
        nonlocal test_times
        test_times += 1
        return text

    ret = _fun(test_data)
    assert ret[0] == test_data and isinstance(ret[1], int)

    @hpydecorator.funName
    def _fun2(text: str, __fun_name__: str):
        return text, __fun_name__

    assert (test_data, "_fun2") == _fun2(test_data)



import os
import random
import time

from . import hpysettings

# 测试数据
setting_file_path = os.path.join(os.getcwd(), "settings_file_dir", str(time.time_ns()))
test_text = random.randint(0, 9999)
test_key = str(random.randint(0, 9999)) + str("test_key")
test_key2 = str(random.randint(0, 9999)) + str("test_key2")
test_value = random.randint(0, 9999)
test_value2 = True
test_value3 = test_text
json_file_name = f"{test_text}.json"
yaml_file_name = f"{test_text}.yaml"
toml_file_name = f"{test_text}.toml"


# 创建文件实例
instance_json_settings_file = hpysettings.load(
    settings_dir_path=setting_file_path,
    settings_file_name=json_file_name.split(".")[0],
    settings_file_format=json_file_name.split(".")[1],
)

instance_toml_settings_file = hpysettings.load(
    settings_dir_path=setting_file_path,
    settings_file_name=toml_file_name.split(".")[0],
    settings_file_format=toml_file_name.split(".")[1],
)

instance_yaml_settings_file = hpysettings.load(
    settings_dir_path=setting_file_path,
    settings_file_name=yaml_file_name.split(".")[0],
    settings_file_format=yaml_file_name.split(".")[1],
)


def test_settings_file():
    """
    测试文件实例

    :return: None
    """
    for instance_settings_file, file_name in (
        (instance_yaml_settings_file, yaml_file_name),
        (instance_toml_settings_file, toml_file_name),
        (instance_json_settings_file, json_file_name),
    ):
        # 测试，add read readAll
        instance_settings_file.add(test_key, test_value)
        assert test_value == instance_settings_file.read(test_key)
        assert {test_key: test_value} == instance_settings_file.readAll()

        # 测试，modify read readAll
        instance_settings_file.modify(test_key, test_value2)
        assert test_value2 == instance_settings_file.read(test_key)
        assert {test_key: test_value2} == instance_settings_file.readAll()

        # 测试，delete add exists
        instance_settings_file.delete(test_key)
        instance_settings_file.add(test_key, test_value3)
        assert instance_settings_file.exists(test_key) is True

        # 测试，setting_file_path
        assert instance_settings_file.setting_file_path == os.path.join(
            setting_file_path, file_name
        )

        # 测试读取不存在的键
        try:
            instance_settings_file.read(test_key2)
            # pytest.fail("读取不存在的键时未能引发KeyError错误")
        except KeyError:
            ...

        # 测试删除不存在的键
        try:
            instance_settings_file.delete(test_key2)
            # pytest.fail("删除不存在的键时未能引发KeyError错误")
        except KeyError:
            ...

        # 测试修改不存在的键
        try:
            instance_settings_file.modify(test_key2, test_value)
            # pytest.fail("修改不存在的键时未能引发KeyError错误")
        except KeyError:
            ...



import tempfile
from random import randint
from typing import IO, Any, AnyStr

import pytest

# # 这样才可以导入上层包哈哈
# sys.path.append(os.path.join(sys.path[0], ".."))
from . import hpycore, output_queue

test_buffer: Any = 0  # 初始化一个变量，用于检测结果
test_reflect: Any = 0  # 初始化一个变量，用于检测结果


class io(IO):
    """模拟打开的文件流"""

    def __init__(self):
        ...

    def write(self, text: AnyStr) -> None:  # type: ignore
        global test_buffer
        test_buffer = text

    def flush(self) -> None:
        global test_reflect
        test_reflect = test_buffer


instance_io = io()  # type: ignore # 创建实例 __enter__和__exit__方法测试用不到就算了

# 测试数据
num = randint(0, 99999)
str_num = str(num)
test_data = str(num) + "test_data"
test_bool = True


def test_hpycore():
    """
    测试hpycore模块

    :return:
    """
    assert (num + 1) == hpycore.addOne(num)

    hpycore.setIoInstance(instance_io)
    assert instance_io == hpycore.getIoInstance()

    hpycore.write(test_data, end="\n")
    assert test_reflect == f"{test_data}\n"

    hpycore.write_without_flush(test_data, end="\n")
    assert test_buffer == f"{test_data}\n"
    hpycore.flush()
    assert test_reflect == f"{test_data}\n"

    hpycore.output(test_data)
    assert output_queue.get() == test_data
    hpycore.output(num)
    assert output_queue.get() == str_num



from pathlib import Path

from setuptools import find_packages, setup

description = "这是一个基于python的高拓展性计算器"

# with open(r"..\README.md", "r", encoding="utf-8") as fh:
#     long_description = fh.read()

THIS_DIR = Path(__file__).resolve().parent
# try:
#     long_description = (THIS_DIR / 'README.md').read_text()
# except FileNotFoundError:
#     long_description = description
long_description = (THIS_DIR / "README.md").read_text(encoding="utf-8")

VERSION = "1.5.0"

setup(
    # 以下为必需参数
    name="hpyculator",
    version=VERSION,
    description=description,
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/HowieHz/hpyculatorPackage",
    author="HowieHz",  # 作者名
    classifiers=[
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.10",
    ],
    install_requires=[
        "rtoml",
        "pyyaml",
    ],
    project_urls={
        "Bug Reports": "https://github.com/HowieHz/hpyculatorPackage/issues",
    },
)



"""用于创建yaml文件对象"""
from __future__ import annotations

from typing import Any

import yaml

from .settings_file_object import SettingsFileObject


class YamlSettingsFileObject(SettingsFileObject):
    """yaml类型的文件对象"""

    def __init__(
        self,
        settings_dir_path: str,
        settings_file_name: str = "settings",
        settings_file_format: str = "yaml",
    ):
        """读取一个文件, 初始化文件对象

        :param settings_dir_path: 设置文件所在的目录
        :type settings_dir_path: str
        :param settings_file_name: 设置文件的名称, 默认为"settings"
        :type settings_file_name: str
        :param settings_file_format: 设置文件的后缀, 默认为"yaml"
        :type settings_file_format: str
        """
        super().__init__(settings_dir_path, settings_file_name, settings_file_format)
        with open(self._settings_file_path, mode="r+", encoding="utf-8") as f:
            if yaml.safe_load(f) is None:
                yaml.dump({}, f)

    def add(self, key: str, value: Any = None) -> YamlSettingsFileObject:
        """添加一项配置

        :param key: 键
        :type key: str
        :param value: 值
        :type value: Any
        :return: 链式调用, 返回本身
        :rtype: YamlSettingsFileObject
        """
        settings_dict = self.readAll()
        with open(self._settings_file_path, mode="w+", encoding="utf-8") as f:
            settings_dict[key] = value
            yaml.dump(settings_dict, f)
        return self

    def read(self, key: str) -> Any:
        """读取一项配置

        :param key: 键
        :type key: str
        :return: 值
        :rtype: Any
        :raises KeyError: 没有这个键
        """
        super().read(key)
        settings_dict = self.readAll()
        return settings_dict[key]

    def readAll(self) -> dict:
        """读取全部的

        :return: 键值对
        :rtype: dict
        """
        with open(self._settings_file_path, mode="r", encoding="utf-8") as f:
            settings_dict = yaml.safe_load(f)
        return settings_dict

    def delete(self, key: str) -> YamlSettingsFileObject:
        """删除一项配置

        :param key: 键
        :type key: str
        :return: 链式调用, 返回本身
        :rtype: YamlSettingsFileObject
        :raises KeyError: 没有这个键
        """
        super().delete(key)
        settings_dict = self.readAll()
        with open(self._settings_file_path, mode="w+", encoding="utf-8") as f:
            settings_dict.pop(key)
            yaml.dump(settings_dict, f)
        return self

    def modify(self, key: str, value: Any) -> YamlSettingsFileObject:
        """修改一项配置

        :param key: 键
        :type key: str
        :param value: 值
        :type value: Any
        :return: 链式调用, 返回本身
        :rtype: YamlSettingsFileObject
        :raises keyError: 没有这个键
        """
        super().modify(key, value)
        settings_dict = self.readAll()
        with open(self._settings_file_path, mode="w+", encoding="utf-8") as f:
            settings_dict[key] = value
            yaml.dump(settings_dict, f)
        return self

    def exists(self, key: str) -> bool:
        """检查一个键是否存在

        :param key: 键
        :type key: str
        :return: 存在为True, 不存在为False
        :rtype: bool
        """
        with open(self._settings_file_path, mode="r", encoding="utf-8") as f:
            is_exists = key in yaml.safe_load(f)
        return is_exists



from . import hpyfunc


def test_hpyfunc():
    test_list = [1, 2, 3, [4, 5, [6, 7]]]
    assert [1, 2, 3, 4, 5, 6, 7] == hpyfunc.flatten(test_list)
    assert [1, 2, 3, 4, 5, [6, 7]] == hpyfunc.flatten_layer(test_list)
    assert [1, 2, 3, 4, 5, 6, 7] == hpyfunc.flatten_no_recursion(test_list)
